"""
Catch Exception decorator for CLI commands.

When any exception is throwed by `InfraRuntimeException` class the console output \
    does not have the stack trace, just the message

#### Usage:

```python
from cdk_organizer.catch_exceptions import catch_exceptions


@catch_exceptions
def myfunction():
    pass
```
"""

import logging
import sys
from functools import wraps
from typing import Callable

logger = logging.getLogger(__name__)


class InfraRuntimeException(Exception):
    """Default managed runtime exception."""


def catch_exceptions(func: Callable) -> Callable:
    """
    Catches and simplifies expected errors thrown by CLI.

    Args:
        func (Callable): The function which may throw exceptions which should be simplified.

    Returns:
        The decorated function.
    """
    @wraps(func)
    def decorated(*args, **kwargs):
        """Invoke function and catches the errors."""
        try:
            result = func(*args, **kwargs)
            return result
        except (InfraRuntimeException) as error:
            logger.error(str(error))
            sys.exit(1)

    return decorated



"""CDK Infra Core Terraform Base Stack."""

from typing import TYPE_CHECKING

from cdktf import S3Backend, TerraformStack

if TYPE_CHECKING:
    from constructs import Construct
    from cdk_organizer.aws.stack_group import StackGroup

from cdk_organizer.decorators.catch_exceptions import catch_exceptions
from cdk_organizer.stack import BaseStack


class Stack(BaseStack, TerraformStack):
    """
    Base class for Terraform CDK Stacks.

    The Terraform Backend is configured to the S3 Bucket and DynamoDB table.

    The S3 Bucket stores the Terraform state files and the DynamoDB table locks the state when it is in use.

    Args:
        scope (Construct): AWS CDK Construct object
        id (str): Stack Id
        stack_group (StackGroup): StackGroup instance
    """

    @catch_exceptions
    def __init__(
        self,
        scope: "Construct",
        id: str,
        stack_group: "StackGroup",
        **kwargs
    ) -> None:
        """Initialize the class."""
        BaseStack.__init__(self, id, stack_group)
        TerraformStack.__init__(self, scope, id, **kwargs)

        backend_config = stack_group.config.get('s3_backend', None)
        if backend_config:
            self.backend_base_args = backend_config
            S3Backend(
                self,
                key=f'{id}/terraform.tfstate',
                **self.backend_base_args,
            )



"""CDK Infra Core AWS Base Stack."""

from typing import TYPE_CHECKING

import aws_cdk as cdk

if TYPE_CHECKING:
    from constructs import Construct
    from cdk_organizer.aws.stack_group import StackGroup

from cdk_organizer.decorators.catch_exceptions import catch_exceptions
from cdk_organizer.stack import BaseStack


class Stack(BaseStack, cdk.Stack):
    """
    Base class for AWS CDK Stacks.

    Args:
        scope (Construct): AWS CDK Construct object
        id (str): Stack Id
        stack_group (StackGroup): StackGroup instance
    """

    @catch_exceptions
    def __init__(
        self,
        scope: "Construct",
        id: str,
        stack_group: "StackGroup",
        **kwargs
    ) -> None:
        """Initialize the class."""
        BaseStack.__init__(self, id, stack_group)
        cdk.Stack.__init__(self, scope, id, env=self.env_props, **kwargs)

        cdk.Tags.of(self).add('Environment', self.env_name.upper())
        if 'tags' in self.config:
            for key, value in self.config['tags'].items():
                cdk.Tags.of(self).add(key, value)

        self.node.add_metadata('cdk:module', self.stack_group.module_name)



"""Logging Module."""

import logging
import warnings

with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=DeprecationWarning)


def setup_logging(module: str, loglevel='INFO') -> logging.Logger:
    """
    Set up logging.

    By default, the python logging module is configured to push logs to stdout as long as \
        their level is at least INFO. The log format is set to `[%(asctime)s] - %(name)s - %(message)s` and \
            the date format is set to `%Y-%m-%d %H:%M:%S`.

    To use the logger configuration add this code block in the CDK main Python script:

    ```python
    from cdk_organizer.logging import setup_logging

    logger = setup_logging(module_name, app.node.try_get_context("loglevel") or "INFO")
    ```

    To use in the modules or stacks

    ```python
    import logging

    logger = logging.getLogger(__name__)
    ```

    Args:
        module (str): python module name `string`
        loglevel (str): log level (`DEBUG`, `INFO`, `WARNING`, `ERROR`, `CRITICAL`).

    Returns:
        A logger instance.
    """
    log_handler = logging.StreamHandler()
    log_handler.setFormatter(_get_formatter())
    logger = logging.getLogger(module)
    for handler in logger.handlers:
        logger.removeHandler(handler)

    logger.addHandler(log_handler)
    logger.setLevel(loglevel)
    return logger


def _get_formatter() -> logging.Formatter:
    """
    Get Logger Formatter `[%(asctime)s] - %(message)s`.

    Returns:
        logger formatter
    """
    fmt = "[%(asctime)s] - %(message)s"
    return logging.Formatter(
        fmt=fmt,
        datefmt="%Y-%m-%d %H:%M:%S"
    )



"""Base CDK Stack module."""

import logging
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from cdk_organizer.aws.stack_group import StackGroup

from cdk_organizer.decorators.catch_exceptions import catch_exceptions


class BaseStack(object):
    """
    Base class for any CDK Stack classes.

    Features:
        - Logging
        - Configuration in dict and dataclass formats
        - Environment properties
        - Utils for generating bucket and general resources names
    """

    @catch_exceptions
    def __init__(
        self,
        stack_name: str,
        stack_group: "StackGroup"
    ) -> None:
        """Initialize the class."""
        self._stack_name = stack_name
        self.stack_group = stack_group
        self.config = stack_group.config

        if hasattr(self.stack_group, 'data'):
            self.data = self.stack_group.data

        self.logger = logging.getLogger(__name__)
        self.env_props = self._get_aws_environment_props(self.config)

    @property
    def env_name(self) -> str:
        """Get the environment name."""
        return self.config.get('env')

    def _get_aws_environment_props(self, config: dict) -> dict:
        """
        Get AWS environment properties from the config YAML file.

        Args:
            config dict: environment config data

        Returns:
            account/region mapping
        """
        env_props = None

        if 'account' in config and 'region' in config:
            env_props = {
                'account': config['account'],
                'region': config['region']
            }

        return env_props

    def get_bucket_name(
        self,
        name: str,
        include_path_naming: bool = True
    ) -> str:
        """
        Generate a bucket name based on following pattern.

        **Pattern**: `{base_naming}-{module_path}-{name}-{region}-{env}`

        The `{base_naming}` is configured in the `config.yaml`, property `baseBucket`.

        Consider the following example:

        **base_naming**: `mycompany`
        **module_path**: `myproject.myapp.www`
        **name**: `spa`
        **region**: `us-east-1`
        **env**: `dev`

        The bucket name will be:

        - `mycompany-myproject-myapp-www-spa-us-east-1-dev`

        Args:
            name (str): string value to be used as the base of the bucket name
            include_path_naming (bool): include the module path in the bucket name

        Returns:
            bucket name
        """
        bucket_name = self.config.get('base_bucket', '')
        if bucket_name != '':
            bucket_name += '-'

        module_name = self.stack_group.normalized_module_name()

        env = self.config.get('env', None)
        region = self.config.get('region', None)

        if include_path_naming:
            bucket_name += f'{module_name}-'

        if name:
            bucket_name += f'{name.lower()}-'

        bucket_name += f"{region}-{env}"
        return bucket_name

    def get_resource_name(
        self,
        name: Optional[str] = None,
        namespace: bool = False,
        database: bool = False,
        use_region: bool = True,
        use_short_region: bool = False,
        ignore_module_path: bool = False
    ) -> str:
        """
        Generate a resource name based on following patterns.

        ### Namespace

        **Pattern**: `{module_path}/{name}/{region}/{env}`

        Consider the following example:

        **module_path**: `myproject.myapp.www`
        **name**: `spa`
        **region**: `us-east-1`
        **env**: `dev`

        The resource name will be:

        - `myproject/myapp/www/spa/us-east-1/dev`

        ### Database

        **Pattern**: `{module_path}_{name}_{env}`

        Consider the following example:

        **module_path**: `myproject.myapp`
        **name**: `raw`
        **env**: `dev`

        The resource name will be:

        - `myproject_myapp_raw_dev`

        Args:
            name (str): string value to be used as the base of the bucket name
            namespace (bool): Use the separator `/` to generate the resource name
            database (bool): Use the separator `_` to generate the resource name
            use_region (bool): Use the region in the resource name, default `true`
            use_short_region (bool): Use the short region in the resource name, default `false`
            ignore_module_path (bool): Ignore the module path in the resource name, default `false`

        Returns:
            resource name
        """
        return self.stack_group.get_resource_name(
            name=name,
            namespace=namespace,
            database=database,
            use_region=use_region,
            use_short_region=use_short_region,
            ignore_module_path=ignore_module_path
        )



"""Base Stack Group module."""

import abc
import dataclasses
import importlib
import inspect
import logging
import os
import sys
from pathlib import Path
from typing import Dict, Generic, Optional, Type, TypeVar, get_args

from cdk_organizer.decorators.catch_exceptions import InfraRuntimeException, catch_exceptions
from cdk_organizer.loaders.config_loader import ConfigLoader
from constructs import IConstruct
from dacite import from_dict

CDK_APP_TYPE = TypeVar('CDK_APP_TYPE', bound=IConstruct)
CDK_CONFIG_TYPE = TypeVar('CDK_CONFIG_TYPE', covariant=True)
CDK_STACK_GROUP_TYPE = TypeVar('CDK_STACK_GROUP_TYPE')


LOGGER = logging.getLogger(__name__)


class StackGroupLoader:
    """
    Loads the stack groups from the stacks directory and synthesizes them into the CDK app.

    This class is passed to all the stack groups and they can use it to load other stacks.

    Args:
        app (CDK_APP_TYPE): The CDK app.

    Attributes:
        app (CDK_APP_TYPE): The CDK app.
        stack_groups (Dict[str, CDK_STACK_GROUP_TYPE]): The stack groups.
    """

    def __init__(self, app: CDK_APP_TYPE) -> None:
        """Stack group loader constructor."""
        self.app = app
        self.stack_groups: Dict[str, CDK_STACK_GROUP_TYPE] = {}
        self._stack_dir = app.node.try_get_context("stacksDirectory") or "stacks"

    def _fullname(self, obj: Type[CDK_STACK_GROUP_TYPE]) -> str:
        """Return the fully qualified name of a class."""
        module = obj.__module__
        if module == 'builtins':
            return obj.__qualname__  # avoid outputs like 'builtins.str'
        return module + '.' + obj.__qualname__

    def synth(self) -> None:
        """Load all the python files from the stacks directory, filters the classes that are `StackGroup` and load the stacks into the CDK app."""
        for file in Path(f'{self._stack_dir}/').rglob("**/*.py"):
            module_name = str(file).replace("/", ".").replace(".py", "")
            spec = importlib.util.spec_from_file_location(module_name, str(file))
            stack_group_module = importlib.util.module_from_spec(spec)
            sys.modules[module_name] = stack_group_module
            spec.loader.exec_module(stack_group_module)

            for _, obj in inspect.getmembers(stack_group_module):
                if not inspect.isabstract(obj) and inspect.isclass(obj) and obj.__base__.__base__ is StackGroup:
                    stack_group_name = self._fullname(obj)
                    if stack_group_name not in self.stack_groups:
                        module_instance = obj(self.app, self)
                        if module_instance.enabled:
                            self.stack_groups[stack_group_name] = module_instance
                            module_instance._load_stacks()

    def resolve_group(self, stack_group_type: Type[CDK_STACK_GROUP_TYPE]) -> CDK_STACK_GROUP_TYPE:
        """
        Resolve a stack group by its type.

        Usage:

        ```python
        self.resolve_group(MyStackGroup)
        ```

        Args:
            stack_group_type (Type[CDK_STACK_GROUP_TYPE]): The stack group type.

        Returns:
            The stack group.
        """
        stack_group_name = self._fullname(stack_group_type)

        if stack_group_name not in self.stack_groups:
            stack_group_instance = stack_group_type(self.app, self)
            self.stack_groups[stack_group_name] = stack_group_instance
            if stack_group_instance.enabled:
                stack_group_instance._load_stacks()

        return self.stack_groups[self._fullname(stack_group_type)]


class StackGroup(
    Generic[CDK_APP_TYPE, CDK_CONFIG_TYPE]
):
    """
    Base Stack Group class.

    This class expects two types as generic arguments:

    - **CDK_APP_TYPE**: The CDK App Type
        - For `aws-cdk` stack group classes, the `app` parameter is the `cdk.App` class.
        - For `cdktf` stack group classes, the `app` parameter is the `cdktf.App` class.

    - **CDK_CONFIG_TYPE**: The Config Dataclass Type

    This class is also a singleton class.

    Usage:
        ### AWS CDK
        ```python
        import aws_cdk as cdk
        from dataclasses import dataclass
        from cdk_organizer.aws.stack_group import StackGroup

        @dataclass
        class MyStackGroupConfig:
            domain: str
            ipv6: bool = True


        class MyStackGroup(StackGroup[MyStackGroupConfig]):
            def _load_stacks(self) -> None:
                MyStack(
                    self.app,
                    self.get_stack_name('my-stack'),
                    stack_group=self,
                    parameters=MyStackParameters(
                        domain=self.data.domain,
                        ipv6=self.data.ipv6
                    )
                )
        ```
        ### Terraform

        ```python
        import cdktf
        from dataclasses import dataclass
        from cdk_organizer.aws.stack_group import StackGroup

        @dataclass
        class MyStackGroupConfig:
            domain: str
            ipv6: bool = True


        class MyStackGroup(StackGroup[cdktf.App, MyStackGroupConfig]):
            def _load_stacks(self) -> None:
                MyStack(
                    self.app,
                    self.get_stack_name('my-stack'),
                    stack_group=self,
                    parameters=MyStackParameters(
                        domain=self.data.domain,
                        ipv6=self.data.ipv6
                    )
                )
        ```
    """

    @catch_exceptions
    def __init__(
        self,
        app: CDK_APP_TYPE,
        loader: StackGroupLoader
    ) -> None:
        """Initialize the Stack Group."""
        self.app = app
        self._loader = loader
        self.dependencies = []
        self.env = os.getenv("CDK_ENV", None) or app.node.try_get_context("env")
        self.region = app.node.try_get_context("region")
        normalized_module_name = self.__module__.replace(".py", "")
        self.module_name = '.'.join(normalized_module_name.split('.')[:-1])
        self.config, self.enabled = ConfigLoader(self.app, self.env, self.region).load_config(self.module_name)

        config_type = self._resolve_config_type()
        if self.enabled and config_type is not None and dataclasses.is_dataclass(config_type):
            self.data = self._get_data(config_type)

    def resolve_group(self, stack_group_type: Type[CDK_STACK_GROUP_TYPE]) -> CDK_STACK_GROUP_TYPE:
        """
        Resolve a dependent stack group instance.

        Args:
            stack_group_type (Type[CDK_STACK_GROUP_TYPE]): The stack group type.

        Returns:
            The stack group.
        """
        stack_group_instance = self._loader.resolve_group(stack_group_type)
        if stack_group_instance not in self.dependencies:
            self.dependencies.append(stack_group_instance)

        return stack_group_instance

    def _resolve_config_type(self) -> Optional[Type[CDK_CONFIG_TYPE]]:
        if hasattr(self.__class__, '__orig_bases__'):
            bases = self.__class__.__orig_bases__
            if isinstance(bases, tuple) and bases:
                type_args = get_args(bases[0])
                if type_args:
                    return type_args[0]

        return None

    def _get_data(self, config_type: Type[CDK_CONFIG_TYPE]) -> CDK_CONFIG_TYPE:
        try:
            return from_dict(
                data_class=config_type,
                data=self.config
            )
        except Exception as e:
            raise InfraRuntimeException(f"Error on parsing config YAML data for stack group '{self.module_name}', {str(e)}")

    def get_stack_name(self, name: Optional[str] = None, ignore_module_path: bool = False) -> str:
        """
        Generate a stack name based on following pattern.

        **Pattern**: `{module_path}-{name}-{region}-{env}`

        Consider the following example:

        **module_path**: `myproject.myapp.www`
        **name**: `spa`
        **region**: `us-east-1`
        **env**: `dev`

        Stack name will be:

        - `myproject-myapp-www-spa-us-east-1-dev`

        Args:
            name (str, optional): string value to be added between the module path and region + env name.
            ignore_module_path (bool, optional): If True, the module path will be ignored. Defaults to False.

        Returns:
            The stack name.
        """
        module_name = self.normalized_module_name()
        stack_name = ''

        env = self.config.get('env', '')
        region = self.config.get('region', 'us-east-1')
        if not ignore_module_path:
            stack_name = f'{module_name}-'

        if name:
            stack_name += f'{name.lower()}-'

        stack_name += f'{region.lower()}-{env.lower()}'
        return stack_name

    def get_resource_name(
        self,
        name: Optional[str] = None,
        namespace: bool = False,
        database: bool = False,
        use_region: bool = True,
        use_short_region: bool = False,
        ignore_module_path: bool = False
    ) -> str:
        """
        Generate a resource name based on following patterns.

        ### Namespace

        **Pattern**: `{module_path}/{name}/{region}/{env}`

        Consider the following example:

        **module_path**: `myproject.myapp.www`
        **name**: `spa`
        **region**: `us-east-1`
        **env**: `dev`

        The resource name will be:

        - `myproject/myapp/www/spa/us-east-1/dev`

        ### Database

        **Pattern**: `{module_path}_{name}_{env}`

        Consider the following example:

        **module_path**: `myproject.myapp`
        **name**: `raw`
        **env**: `dev`

        The resource name will be:

        - `myproject_myapp_raw_dev`

        Args:
            name (str): string value to be used as the base of the bucket name
            namespace (bool): Use the separator `/` to generate the resource name
            database (bool): Use the separator `_` to generate the resource name
            use_region (bool): Use the region in the resource name, default `true`
            use_short_region (bool): Use the short region in the resource name, default `false`
            ignore_module_path (bool): Ignore the module path in the resource name, default `false`

        Returns:
            resource name
        """
        separator = '-'
        resource_name = ''
        if namespace:
            separator = '/'
        elif database:
            separator = '_'

        module_name = self.normalized_module_name(separator)
        if not ignore_module_path:
            resource_name = f'{module_name.lower()}{separator}'

        env = self.config.get('env')
        region = self.config.get('region')

        if name:
            resource_name += f'{name.lower()}{separator}'

        if database:
            resource_name += f'{env}'
            resource_name = resource_name.replace('-', separator)
        else:
            if use_region:
                if use_short_region:
                    region = self.config.get('short_region')
                resource_name += f'{region}{separator}{env}'
            else:
                resource_name += f'{env}'

        return resource_name

    def normalized_module_name(self, separator: str = '-') -> str:
        """
        Normalize the module name by removing the `stacks` prefix and replacing all dots with a separator.

        Args:
            separator (str): The separator to use, default value: `-`

        Returns:
            The normalized module name.
        """
        parts = self.module_name.split('.')

        ignore_stacks_prefix = self.app.node.try_get_context('ignoreStacksPrefix') or False
        if not ignore_stacks_prefix:
            parts[0] = parts[0].replace('stacks', '')
            if (parts[0].endswith('_')):
                parts[0] = parts[0][:-1]
        else:
            parts.pop(0)

        return separator.join(parts).lower().replace('_', '-')

    @abc.abstractmethod
    def _load_stacks(self) -> None:
        """Load stacks."""



"""Terraform CDK Stack Group Class."""

from typing import TypeVar

from cdk_organizer.stack_group import StackGroup as BaseStackGroup
from cdk_organizer.stack_group import StackGroupLoader
from cdktf import App

CDK_CONFIG_TYPE = TypeVar('CDK_CONFIG_TYPE', covariant=True)


class StackGroup(BaseStackGroup[App, CDK_CONFIG_TYPE]):
    """Stack Group for Terraform CDK."""

    def __init__(self, app: App, loader: StackGroupLoader) -> None:
        """Stack group constructor."""
        super().__init__(app, loader)



"""Include YAML tag constructor."""

import json
import os
from typing import IO, Any

import yaml


def custom_compose_document(self):
    """Override the default compose_document method, and removes the `self.anchors = {}` line."""
    self.get_event()
    node = self.compose_node(None, None)
    self.get_event()
    return node


yaml.SafeLoader.compose_document = custom_compose_document


def yaml_path_loader(path: str) -> yaml.SafeLoader:
    """
    YAML Path Loader.

    Example:

    `a.yaml`

    ```yaml
    b: !include b.yaml
    ```

    `b.yaml`

    ```yaml
    key: value
    ```

    Args:
        path (str): YAML file path

    Returns:
        yaml.SafeLoader: YAML Loader
    """
    class Loader(yaml.SafeLoader):
        def __init__(self, stream: IO) -> None:
            self._root = os.path.dirname(path)
            super().__init__(stream)

    def construct_include(loader: Loader, node: yaml.Node) -> Any:
        filename = os.path.abspath(os.path.join(loader._root, loader.construct_scalar(node)))
        extension = os.path.splitext(filename)[1].lstrip('.')

        with open(filename, 'r') as f:
            if extension in ('yaml', 'yml'):
                included_loader = yaml.SafeLoader(f.read())
                included_loader.anchors = loader.anchors

                return included_loader.get_data()
            elif extension in ('json', ):
                return json.load(f)
            else:
                return f.read()

    yaml.add_constructor('!include', construct_include, Loader)
    return Loader



"""
Config Loader.

All the config files are located in the `config` folder and must match with the stacks folder.

### Example:

If stack group is located at `stacks.cloe.cdc.stacks` the config file \
    must be at `config/<env>/<region>/cloe/cdc/config.yaml`.

The files are also loaded recursively and all the propertires are merged together.

### Example:

Consider the following config structure:

```text
.
+-- config/
| +-- config.yaml (`prop1: value1`)
| +-- dev/
|   +-- config.yaml (`prop2: value2`)
|   +-- us-east-/
|       +-- config.yaml (`prop3: value3`)
|       +-- app1/
|           +-- config.yaml (`prop4: value4`)
```

The config object for the `dev/us-east-/app1` stack will be:

```yaml
prop1: value1
prop2: value2
prop3: value3
prop4: value4
```

> **Note**: If the property name conflicts, the higher priority config file will \
    override the lower priority config file.

"""

import os
from fnmatch import fnmatch
from pathlib import Path
from typing import Tuple, TypeVar

import yaml
from cdk_organizer.miscellaneous.yaml_tags.include_yaml import yaml_path_loader
from constructs import IConstruct

CDK_APP_TYPE = TypeVar('CDK_APP_TYPE', bound=IConstruct)


class ConfigLoader(object):
    """Configuration Loader from YAML file to dict object based on the CDK Stack Groups."""

    def __init__(self, app: CDK_APP_TYPE, env: str, region: str) -> None:
        """
        Initialize the Configuration Loader.

        Args:
            app (CDK_APP_TYPE): CDK App instance
            env (str): environment name
            region (str): region name
        """
        super().__init__()

        self.app = app
        self.env = env
        self.region = region
        self._stack_dir = app.node.try_get_context("stacksDirectory") or "stacks"
        self._config_dir = app.node.try_get_context("configDirectory") or "config"

    def merge_dict(self, dict1: dict, dict2: dict, path=None) -> dict:
        """
        Merge two dict objects.

        Args:
            dict1 (dict): first dict object
            dict2 (dict): second dict object
            path (list): dict property path
        Returns:
            dict merged dict object
        """
        if path is None:
            path = []
        for key in dict2:
            if key in dict1:
                if isinstance(dict1[key], dict) and isinstance(dict2[key], dict):
                    self.merge_dict(dict1[key], dict2[key], path + [str(key)])
                elif dict1[key] == dict2[key]:
                    pass  # same leaf value
                else:
                    dict1[key] = dict2[key]
            else:
                dict1[key] = dict2[key]
        return dict1

    def _load_config_recursive(self, path: Path, config: dict = {}) -> dict:
        if not path.exists():
            return self._load_config_recursive(path.parent, config)
        else:
            for filename in os.listdir(path):
                if fnmatch(filename, "*.yaml"):
                    with open(os.path.join(path, filename), 'r') as file:
                        file_data = yaml.load(file.read(), yaml_path_loader(os.path.join(path, filename)))

                    config = self.merge_dict(file_data or {}, config)

            if str(path) != ".":
                config = self._load_config_recursive(path.parent, config)

            return config

    def load_config(self, module: str) -> Tuple[dict, bool]:
        """
        Load the configuration from the YAML file to the dict object.

        Args:
            module (str): module name

        Returns:
            configuration object and if the module config exists.
        """
        module_path = Path(module.replace(".", "/"))
        module_folder = module_path
        ignore_parts = self._stack_dir.count("/") + 1
        if not module_path.is_dir():
            module_folder = module_path.parent
        module_folder = Path(f'{self._config_dir}/{self.env}/{self.region}').joinpath(Path(*module_folder.parts[ignore_parts:]))
        module_config = module_folder.joinpath("config.yaml")

        config = self._load_config_recursive(module_folder)

        return config, module_config.exists()



"""Merge YAML tag constructor."""

from typing import Any, List

import yaml


def construct_merge(loader: yaml.SafeLoader, node: yaml.Node) -> List[Any]:
    """
    YAML !merge tag constructor.

    This tag is used to merge arrays
    Example:
        YAML file:
        ```yaml
        a: !merge
        - 1
        - 2
        - - 3
        - 4
        ```
        Output:
        ```json
        {
            "a": [
                1,
                2,
                3,
                4
            ]
        }
        ```

    Args:
        loader (Loader): YAML custom loader instance
        node (yaml.nodes.SequenceNode): YAML node
    Returns:
        A merge list of items
    """
    if not isinstance(node, yaml.nodes.SequenceNode):
        raise yaml.constructor.ConstructorError(
            None, None, f'expected a sequence node, but found {node.id}', node.start_mark)

    values = loader.construct_sequence(node)

    result = []
    for value in values:
        if isinstance(value, list):
            result += value
        else:
            result.append(value)

    return result
